import collections
import struct
import math
import time

thetime = 0

def timer(cmd='start'):
    """ Counts elapsed time between start and stop events. 
        Useful for tracking render time. """
    global thetime
    if cmd == 'start':
        thetime = time.time()

        print 'Started render at timestamp', thetime
        return thetime 
    elif cmd == 'stop':
        thetime = time.time() - thetime
        themin = int(thetime) / 60
        thesec = thetime - (themin * 60)
        print 'Render time:', themin, 'min', thesec, 'sec'
        return thetime


class Frame(collections.MutableSequence):
    def __init__(self, samples=None, channels=2):
        super(Frame, self).__init__()

        if samples is None:
            samples = []

        if isinstance(samples, int) or isinstance(samples, float):
            self.samples = [ int(samples) ] * channels
        else:
            self.samples = samples

        self.channels = channels

    def __str__(self):
        def pack(s):
            s = -32768 if s < -32768 else s
            s = 32767 if s > 32767 else s

            return struct.pack('<h', s)

        return ''.join([ pack(sample) for sample in self.samples ])

    def __len__(self):
        return len(self.samples)

    def __getitem__(self, index):
        return self.samples[index]

    def __setitem__(self, index, value):
        self.samples[index] = value

    def __delitem__(self, index):
        return self.samples.pop(index)

    def insert(self, index, value):
        self.samples.insert(index, value)

    def __contains__(self, key):
        for sample in samples:
            if sample == key: return True

        return False

    def __add__(self, value):
        if isinstance(value, int) or isinstance(value, float):
            return Frame([ sample + int(value) for sample in self.samples ])
        elif isinstance(value, Frame):
            return Frame([ sample + value[i] for i, sample in enumerate(self.samples) ])

    def __sub__(self, value):
        if isinstance(value, int) or isinstance(value, float):
            return Frame([ sample - int(value) for sample in self.samples ])
        elif isinstance(value, Frame):
            return Frame([ sample - value[i] for i, sample in enumerate(self.samples) ])

    def __rsub__(self, value):
        if isinstance(value, int) or isinstance(value, float):
            return Frame([ int(value) - sample for sample in self.samples ])
        elif isinstance(value, Frame):
            return Frame([ value[i] - sample for i, sample in enumerate(self.samples) ])

    def __mul__(self, value):
        if isinstance(value, int) or isinstance(value, float):
            return Frame([ int(sample * value) for sample in self.samples ])
        elif isinstance(value, Frame):
            return Frame([ sample * value[i] for i, sample in enumerate(self.samples) ])


class Sound(collections.MutableSequence):
    def __init__(self, frames=None, channels=2):
        super(Sound, self).__init__()

        self.frames = []

        if isinstance(frames, list):
            for i, frame in enumerate(frames):
                if isinstance(frame, int) or isinstance(frame, float):
                    self.frames += [ Frame(int(frame), channels) ]
                elif isinstance(frame, tuple):
                    self.frames += [ Frame([f for f in frame]) ]
                elif isinstance(frame, Frame):
                    self.frames += [ frame ]

        self.channels = channels

    def __len__(self):
        return len(self.frames)

    def __str__(self):
        return ''.join([ str(frame) for frame in self.frames ])

    def __contains__(self, key):
        exists = 0
        for frame in self.frames:
            if frame.contains(key):
                exists += 1

        return exists > 0

    def __getitem__(self, index):
        return self.frames[index]

    def __setitem__(self, index, value):
        if isinstance(value, Frame) and len(value) == self.channels:
            self.frames[index] = value
        elif isinstance(value, int) or isinstance(value, float):
            self.frames[index] = Frame(int(value), self.channels)

    def __delitem__(self, value):
        if isinstance(value, int):
            return self.frames.pop(value)
        elif isinstance(value, tuple):
            return self.frames.remove(value)

    def insert(self, index, value):
        self.frames.insert(index, value)

    def append(self, value):
        if isinstance(value, Sound) or isinstance(value, list):
            self.frames += [ value ]
        elif isinstance(value, int) or isinstance(value, float):
            self.frames += [ Frame(value) ]
        else:
            print 'foo'

    def loop(self, loops):
        self.frames = self.frames * loops

    def __iadd(self, value):
        return self.__add__(value)

    def __add__(self, value):
        if isinstance(value, int):
            return Sound([ tuple([ chan + value for chan in frame ]) for frame in self.frames ])
        elif isinstance(value, Sound):
            return Sound([ frame + value[i] for i, frame in enumerate(self.frames) ])

    def __radd__(self, value):
        return self.__add__(value)

    def __sub__(self, value):
        if isinstance(value, int):
            return Sound([ frame - int(value) for frame in self.frames ])
        elif isinstance(value, Sound):
            return Sound([ frame - value[i] for i, frame in enumerate(self.frames) ])
    
    def __rsub__(self, value):
        if isinstance(value, int):
            return Sound([ int(value) - frame for frame in self.frames ])
        elif isinstance(value, Sound):
            return Sound([ value[i] - frame for i, frame in enumerate(self.frames) ])

    def __mul__(self, value):
        if isinstance(value, int) or isinstance(value, float):
            return Sound([ frame * value for frame in self.frames ])
        elif isinstance(value, Sound):
            return Sound([ frame * value[i] for i, frame in enumerate(self.frames) ])

    def __rmul__(self, value):
        return self.__mul__(value)

if __name__ == '__main__':

    timer('start')

################################################################################
## Writing sounds
################################################################################

# Sound literals with tuple shorthand
    asnd = Sound([ (200,  201), # first stereo frame
                   (-200,-201), 
                   (200,  201), 
                   (-200,-201) # fourth stereo frame...
                ])

# The longhand way - shorthand expands to this internally
    bsnd = Sound([ Frame([300,  301]), 
                   Frame([-300,-301]), 
                   Frame([300,  301]), 
                   Frame([-300,-301])
                ])
# Sound literals with list shorthand csnd = Sound([200, 201, 202, 203])

# Sound literals with mixed list and tuple shorthand
    dsnd = Sound([200, 201, 202, (203, 204)])

# Sound literals with mixed list shorthand and longhand
    esnd = Sound([200, 201, 202, Frame([203, 204])])


################################################################################
## Sound math 
################################################################################

    asnd = asnd + bsnd
    asnd = asnd + 100
    asnd = asnd - bsnd
    asnd = asnd - 400 
    asnd = 50000 - asnd 
    asnd = asnd * 0.6

    asnd[3] = asnd[0]
    asnd[1] = 300

    asnd.pop(2)

################################################################################
## Synthesis
################################################################################

    import wave

    def sines(hz=440.0, length=44100):
        frames = []
        for f in range(length):
            frames += [ int(math.sin(((f % int(hz)) / float(hz)) * math.pi * 2.0) * 32767.0) ]

        return frames

    snd1 = Sound(sines(200.0, 44100 * 3))
    snd2 = Sound(sines(2.0, 44100 * 3))

    snd = snd1 * snd2

    #snd.loop(2)

    timer('stop')

    soundfile = wave.open('renders/snd.wav', 'w')
    soundfile.setparams([2, 2, 44100, 0, 'NONE', 'not_compressed'])
    soundfile.writeframes(str(snd))
    soundfile.close()
